% PIE algorithm to analyze SFG (From PIE Feurer)
% Julien Réhault
% April 2020
% units fs, PHz, um
clear all;
%% constants
c = 0.2998; % speed of light [um/fs]
Lambda_NIR=799; % central wavelength in nm of NIR narrwoband pulse
fmin=2800;%*c/10+c/Lambda_NIR*1e6;
fmax=3020;%*c/10+c/Lambda_NIR*1e6;  % for plots from cm-1 to THz
%% choose data
datanum=30;  % PTY_datanumdata number like in lab book (steps of 200fs)
taun=7; % number of delays to use for reconstruction
%sample_angle = 15 ; %
%% time and frequency vector
tn = 2^12;  % 2^12 is 4096 points
tmax =8e3;  % 10e3 is 10ps
dt = 2*tmax/tn;
t = dt*linspace(-tn/2,tn/2-1,tn);
df = 1/(tn*dt)/c*1e4;  % in cm-1
f = df*linspace(-tn/2,tn/2-1,tn);
w=2*pi*f*0.02998;  % in THz


%% function for PIE


%% Load data and plot
[Isim,delta_t]=LoadPTY(datanum,tn,f,Lambda_NIR,fmax,fmin,0);
step=2; % multiple of step taken in experiment (200fs*2 =400fs here)
start=38; % first delay to use from the whole data (a bit before time zero)
jj=0;
for ii=start:step:step*taun+start-1
    jj=jj+1;
Idat(:,jj)=Isim(:,ii);  % Idat contains only taun spectra separated by step*timestep of data
end;
taun=jj;
%% array of taun time delays (odd number)
dtau = step*0.200; % time delay spacing in fs, the data was taken with 200fs steps
if taun/2 == round(taun/2), taun = taun-1; end;
tau = dtau*linspace((taun-1)/2,-(taun-1)/2,taun);
taumin = min(tau);
taumax = max(tau);
step=linspace(1,taun,taun)
%% Probe shape
Ptau = zeros(taun,tn);
p=1;  % to use the FP pulse
switch p   % for testing different pulse shapes
    case 0
for ii = 1:taun
% assymetric pulse a : rise, b : fall
a =0.3*10e-3;  % rise
b=2200; %(fall time, fs)
shift_t=2000;
Probe_t = (heaviside(-t-tau(ii)-1-b+shift_t).*exp(-a^2*(t+tau(ii)+b-shift_t).^2)+(heaviside(t+tau(ii)+1+b-shift_t).*exp(-(t+tau(ii)+b-shift_t)/b)));
Ptau(ii,:) =Probe_t;
end;
    case 1  % Fabry-Perot pulse
c=0.0003;  % im mm/fs
L=0.0113; % in mm from spec
RR=0.97;  % reflectivity from spec
tauo=L/c;
tau_pulse=97;  % pulse duration for our laser
E=zeros(tn,1);
shift=5500;  % just a time shift 
for i=0:200  % nb reflections to consider in the FP
E = E+exp(-(4*log(2).*(t'+shift-2*tauo*i)/tau_pulse).^2)*RR^(2*i)*(1-RR);
end
Probe_t=sqrt(round(E*100000)./(100000*max(E)))';  % just rounds to 1e-5 
end;
%% windows 
window1=exp(-((t-1100)/5000).^60).*exp(-((t+5000)/6000).^20);
window2=0; % relevant only when updating also the probe (ePIE method) probe
Pwold = fftshift(fft(fftshift(Probe_t)));
%% reconstruction
Iguess =  zeros(1,tn);  % initial guess, can be used also : the probe pulse centered correctly
Otrec=Iguess;
ij = 0;
nob = 50; % number of iterations, convergence can be faster with a better initial guess
err_Ot = zeros(1,nob);
lnerr = 0;
Ptrec=Probe_t;
tau=tau-2.8;
while (ij < nob)
    ij = ij+1;
    if mod(ij,10) == 0 , ij
    end;
    %% function for ePIE, last 3 parameters relevant only for ePIE not for basic PIE
    [Otrec,Ptrec]= PIE_basic(t,f,Idat,Ptrec,tau,1,1e-5,0.1,Otrec,10,window1,window2,fmin,fmax,0,3,0); % see function for details 
    % plot results
    htrec = Otrec/max(abs(Otrec));
    htrec_abs = abs(htrec);
    htrec_ph = unwrap(angle(htrec));
    htrec_ph = htrec_ph-htrec_ph(tn/2+1);
    htrec = htrec_abs.*exp(1i*htrec_ph);
end;
%% reconstructed spectrogram
Irec = zeros(tn,taun);
for ii = 1:taun
    Pwold = fftshift(fft(fftshift(Ptrec)));
    Ptshift = fftshift(ifft(fftshift(Pwold.*exp(1i*(w)*tau(ii)))));
    ht = Ptshift.*Otrec;
    hw = fftshift(fft(fftshift(ht)));
    Irec(:,ii) = abs(hw').^2;
end;
Irec = Irec/max(max(Irec));Irec=((Irec-min(min(Irec)))); % normalize
Isim = Isim/max(max(Isim));Isim=((Isim-min(min(Isim)))); % normalize
%% plot data
close all;f1=figure(1);
%% images in Fig2 a)
p1=subplot(4,2,1)
imagesc(step,f,Idat/max(max(Idat(:,:))));
colorbar;
axis xy;
ylim([fmin fmax]);
set(gca,'Xticklabel',[]);
colormap(jet);
txt='Experimental';
text(5.4,3000,txt,'Color','white')

p2=subplot(4,2,3)
imagesc(step,f,Irec/max(max(Irec(:,:))));
colorbar;
axis xy;
ylim([fmin fmax]);
xlabel('delay #');
yl=ylabel('frequency [cm^-^1]');
yl.Position(1) =-0.3;
yl.Position(2) =3040;
colormap(jet);
txt='Reconstructed';
text(5.4,3000,txt,'Color','white');
set(p1, 'Position', [0.1300    0.7524    0.335   0.1726]);
set(p2, 'Position', [ 0.1300    0.5634    0.335    0.1726]);

%% Fig 2 b)
p3=subplot(4,2,[2 4])
plot(t/1000,real(htrec),'LineWidth',2);
hold on;
plot(t/1000,abs(circshift(Ptrec,500)),'--','Color','r');
plot(t/1000,window1,'--','Color',[0 0 0]+0.3);
hold off;
xlabel('time [ps]');
ylabel('Amplitude(a.u.)');
xlim([-4 6])
ylim([-1.05 1.05])
set(gca,'position',[0.5703    0.5634    0.335    0.3616])

%% Fig2c)
subplot(4,2,[5 7])
httrec=circshift(htrec,870);  % shift to make the phase zero at max
spectrum=fftshift(fft(fftshift(httrec)));
[M I]=max(abs(spectrum(3*tn/4:tn)));
I=I+3*tn/4-1;  % just a phase shift (arbitrary, no reference anyway)
phaseatmax=angle(spectrum(I));
spectrum=spectrum*exp(-j*(phaseatmax));
yyaxis left
set(gca,'ycolor','black') 
plot(f,abs(spectrum)/max(abs(spectrum)),'LineWidth',2,'Color','b');
hold on;
ylim([-0.21 1.05]);ylabel('Amplitude (a.u.)','Color','black');yyaxis right;yl=ylabel('Phase','rotation',-90);
annotation('arrow', [.25 .29], [.26 .26],'Color','r','LineWidth',2,'Headstyle','vback3');
annotation('arrow', [.22 .18], [.37 .37],'Color','b','LineWidth',2,'Headstyle','vback3');
plot(f,angle(spectrum),'LineWidth',2,'Color','r');
ylim([-0.4 2])
set(gca,'YColor','black');
yl.Position(1) =3047;
yl.Position(2) =0.5;
yyaxis left
plot(f,sqrt(Idat(:,1)/max(max(Idat(:,1)))),'--','LineWidth',1,'Color','black'); % square root od spectrum at delay 1
hold off;
xlabel('frequency [cm^-^1]');
xlim([fmin fmax])

% calculating the NRB subtracted spectrum requires fitting of the NRB
 spectrum=fftshift(fft(fftshift(httrec)));
[M I]=max(abs(spectrum(3*tn/4:tn)));
I=I+3*tn/4-1;
phaseatmax=angle(spectrum(I));
spectrum=spectrum*exp(-j*(phaseatmax));
% yyaxis left
% set(gca,'ycolor','black') 
tt_cplx=spectrum/max(abs(spectrum));
tt=abs(tt_cplx)/max(abs(tt_cplx));
tt_phase=angle(tt_cplx);
% we use only specific points from the spectra to fit NRb
NRB=cat(2,tt(:,1:5),tt(tn/2:tn/2+100),tt(:,3364:3410),tt(:,3431:3432),tt(:,3480:3520),tt(:,tn-100:tn));
f2=cat(2,f(:,1:5),f(tn/2:tn/2+100),f(:,3364:3410),f(:,3431:3432),f(:,3480:3520),f(:,tn-100:tn));
NRB_fitdata=interp1(f2,NRB,f,'PCHIP');%NRB_fit created from cftools
NRB_phase1=cat(2,tt_phase(:,3380:3410),tt_phase(:,3435:3450),tt_phase(:,3480:3490));
f3=cat(2,f(:,3380:3410),f(:,3435:3450),f(:,3480:3490));
NRB_phase=interp1(f3,NRB_phase1,f,'PCHIP');
NRB_cplx=NRB_fitdata.*exp(j.*NRB_phase);

%% plot NRB subtracted signal
subplot(4,2,[6 8]);
plot(f,abs(tt_cplx-1.1*NRB_cplx)/max(abs(tt_cplx-1.1*NRB_cplx))/0.632,'Color','b'); % plots the amplitude
hold on;
plot(f,sqrt(Idat(:,4)/max(Idat(:,4))),'--','Color','black'); % square root of spectrum spectrum at delay 4
xlim([fmin fmax]);xlabel('frequency [cm^-^1]');
ylabel('Amplitude a.u.');ylim([-0.6 1.05]);
hold on;
yyaxis right
% plots phase only were amplitude is significant >3% of max
indices=find((abs(tt_cplx-1.1*NRB_cplx).^2)/0.4073<0.03);
phase_small=angle((tt_cplx-1.1*NRB_cplx)*exp(j*pi));
phase_small(indices)=NaN;
plot(f,phase_small,'Color','r') ; 
yl=ylabel('Phase','rotation',-90);ylim([-1.2 2]);yl.Position(1) =3047;yl.Position(2) =0.5;
annotation('arrow', [.82 .86], [.22 .22],'units','normalized','Color','r','LineWidth',2,'Headstyle','vback3');
annotation('arrow', [.66 .62], [.37 .37],'Color','b','LineWidth',2,'Headstyle','vback3');
hold off;

%% 
set(findall(gcf,'-property','FontSize'),'FontSize',7);
set(findall(gcf,'-property','Linewidth'),'Linewidth',1);
set(findall(gcf,'-property','YColor'),'YColor','black');
set(findall(gcf,'-property','FontName'),'FontName','Arial')
%set(findall(gcf,'-property','FontWeight'),'FontWeight','bold');
f1.Units='inches';f1.Position=[-20 5 9 3.75];
saveas(gcf,'fig_2_V6','epsc');


%% function to load data
function [Isim,dt]= LoadPTY(datanum1,tn,f,Lambda_NIR,fmax,fmin,debug)
% function to load ptychographic data saved from SFG setup
% datanum1 : number after PTY_
% tn is ths size
% frequency axis
% lambda_NIR= 800nm, center wavelength in nm;
% fmax &fmin : max freq in cm-1 to make a window
% debug : 0 or 1 for plots or to check 
%% constants
% speed of light [um/fs]
c = 0.2998;

dataname=strcat('PTY_0',num2str(datanum1));
Rawdataname=strcat(dataname,".csv");
data_xaxis=strcat(dataname,"_xaxis.csv");
data_taxis=strcat(dataname,"_delays.csv");

% check delays and scan length in ps
taxis=importdata(data_taxis,'\t');
[dummy taun]=size(taxis);
tscan=(max(taxis)-min(taxis))/c; % in ps
dtscan=tscan/(taun-1); % in ps

xaxis=importdata(data_xaxis,'\t');
for ii=1:1024  % there are 1024 pixels
    xaxis(ii)=1/xaxis(ii)*1e7-1/Lambda_NIR*1e7;           
end;
 % for correct interpolation later
xaxis=cat(2,-xaxis,fliplr(xaxis)); % creates te negative frequencies
xaxis(1)=-10e3;
xaxis(end)=10e3; 

%imports data
Dataraw=importdata(Rawdataname)';
% baseline calculated from data outside window defined by fmax and fmin

Baseline=min(min(Dataraw(1:max(find(fmin<xaxis)-1024),:)));
Data=zeros(1024,taun);
jj=0;
for ii=1:taun
    jj=jj+1;
Data(:,jj)=Dataraw(:,ii)-Baseline;
end
Data=cat(1,Data,fliplr(Data')');

%% window
fc=(fmax+fmin)/2;
df=(fmax-fmin)/2;
f_window=(exp(-((f-fc)/df).^20)+exp(-((f+fc)/df).^20));

for ii=1:taun
newdat(:,taun-ii+1)=(interp1(xaxis,Data(:,ii),f,'linear','extrap')-1).*f_window;
end

Isim=newdat;
dt=dtscan;

if debug==1
figure(1);
plot(f,newdat(:,1:taun));
xlim([fmin fmax])
hold on;
scale =max(max(newdat));
plot(f,f_window*scale);
hold off;
% figure(2);
% plot(taxis,mean(newdat))
figure(3);
plot(f,newdat(:,:))
xlim([fmin fmax])
end;

end
%% function for PIE


function [Otrec,Ptrec] = PIE_basic(t,f,I,Ptrec,tau,beta,alpha,beta_P,Otrec,noi,window1,window2,fmin,fmax,epie,constraint,debug)
% basic PIE reconstruction algorithm
% Thomas Feurer April 2015
% modified by Julien Réhault April 2020 foe ePIE and windowing
% units fs, PHz, um
% t: time vector
% f: frequency vector
% I: spectrogram (size = [fn,taun])
% Ptau: array of time delayed probes (size = [taun,fn])
% beta: reconstruction constant (typically 0.1 ... 2.0)
% alpha: reconstruction constant (typically 1e-4)
% Otrec: initial guess for object function
% noi: number of iterations
%window1 : window for object
% window2 : window for probe
% epie= 0, noe epie, =1 -> epie
% constraint = 0: no constraint
%            = 1: pure amplitude probe
%            = 2: pure phase probe
%            = 3: pure real probe
%            = 4: pure imaginary probe
% debug = 0: no plots generated
%       = 1: one time delay per iteration is shown
%       = 2: all time delays are displayed


%% dimensions
[fn,taun] = size(I);
w=2*pi*f*0.02998; % in THz
fc=(fmax+fmin)/2;
df=(fmax-fmin)/2;
f_window=(exp(-((f-fc)/df).^20)+exp(-((f+fc)/df).^20));

%% reconstruction loop
for ii = 1:noi
    for ik = 1:taun
        % reverse direction at every second iteration
        ij = ik;
        if mod(ii,2) == 0, ij = taun-ik+1; end;
        
        % measured spectrum for actual time delay tau(ij)
        Iw = I(:,ij)';

        % store old object
        Otold = Otrec;
         Ptold = Ptrec;
         
        % UPDATE OBJECT
        % time delayed probe pulse
        %Ptshift = Ptau(ij,:);
        Pwold = fftshift(fft(fftshift(Ptrec)));
        Ptshift = fftshift(ifft(fftshift(Pwold.*exp(1i*(w)*tau(ij)))));
           
        % Fourier transform of product field
        Gtold = Ptshift.*Otold;
        Gwold = fftshift(fft(fftshift(Gtold)));

        % replace modulus with measurement
        Gwnew = sqrt(Iw).*exp(1i*angle(Gwold));%.*f_window;

        % inverse Fourier transformation
        Gtnew = fftshift(ifft(fftshift(Gwnew)));
               
        % difference
        DeltaG = Gtnew-Gtold;
        
        % update object function
        Oupdate = abs(Ptshift)./max(abs(Ptshift)).*conj(Ptshift)./(abs(Ptshift).^2+alpha); % time 
        Otrec = Otold+beta*Oupdate.*DeltaG;
        Otrec= Otrec.*window1;
      
        if mod(ii,5) == 0 && epie == 1, 
          %% UPDATE PROBE PULSE  (for ePIE only)
        % time delayed object
        Owold = fftshift(fft(fftshift(Otold)));
        Otshift = fftshift(ifft(fftshift(Owold.*exp(-1i*w*tau(ik)))));

        % Fourier transform of product field
        Gtold = Ptold.*Otshift;
        Gwold = fftshift(fft(fftshift(Gtold))).*window2;

        % replace modulus with measurement
        Gwnew = sqrt(Iw).*exp(1i*angle(Gwold)).*window1;

        % inverse Fourier transformation
        Gtnew = fftshift(ifft(fftshift(Gwnew)));
        
        % difference
        DeltaG = Gtnew-Gtold;

        % update probe pulse
        Pupdate = conj(Otshift)./max(abs(Otshift).^2);
        Ptrec = Ptold+beta_P*Pupdate.*DeltaG;
        
        % apply constraint to probe pulse
        switch constraint
            case 1
                Ptrec = abs(Ptrec);
            case 2
                Ptrec = exp(1i*angle(Ptrec));
            case 3
                Ptrec = sgolayfilt(real(Ptrec),1,5);
            case 4
                Ptrec = imag(Ptrec);
        end;  
        
        end;
        
        % plots
        if (debug == 1  && ij == (taun+1)/2) || (debug == 2)
            ht = Otrec/max(abs(Otrec));
            ht_ph = angle(ht);
            ht_ph = ht_ph-ht_ph(fn/2+1);
            ht = abs(ht).*exp(1i*ht_ph);
            ht_abs = abs(ht);
            ht_ph = unwrap(angle(ht));
            ht_ph(ht_abs<0.01) = NaN;

            figure(1);
            clf;
            subplot(4,1,1),
            plot(t,abs(Ptshift),t,window1,t,unwrap(angle(Ptshift)),'-r');
            axis tight;
            ylim([0 1])
            title(['iteration ' num2str(ii,4)]);
            xlabel('time [fs]');
            
            subplot(4,1,2),
            plot(f,Iw,'-b'); 
            axis tight;
            title('measured spectrum');
            xlabel('frequency [PHz]');
            xlim([fmin fmax]);
            %ylim([0 8000])
    
            subplot(4,1,3), 
            plot(t,ht_abs,'.r');
            axis tight;
            title('amplitude');
            xlabel('time [fs]');
    
            subplot(4,1,4),
            plot(t,ht_ph,'.r'); 
            axis tight;    
            title('phase');
            xlabel('time [T]');
            drawnow;
            pause(0.05);
        end;   
    end;
end;
end
